import sys
import GPUtil
import os
from core.exception import *


gpus = GPUtil.getGPUs()


def gpu_id():
    if sys.platform == 'win32':
        for gpu in gpus:
            return gpu.id
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        data = os.popen('lshw -C display').read()
        for line in str(data).splitlines():
            if 'physical id' in line:
                return line.replace('physical id: ', '').replace('       ', '')
    else:
        return unsupported_exception()


def gpu_name():
    if sys.platform == 'win32':
        for gpu in gpus:
            return gpu.name
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        data = os.popen('lshw -C display').read()
        for line in str(data).splitlines():
            if 'product: ' in line:
                return line.replace('product: ', '').replace('       ', '')
    else:
        return unsupported_exception()


def gpu_load():
    if sys.platform == 'win32':
        for gpu in gpus:
            return f"{gpu.load*100}%"
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return feature_not_implemented_yet()
    else:
        return unsupported_exception()


def gpu_free_memory():
    if sys.platform == 'win32':
        for gpu in gpus:
            return f"{gpu.memoryFree}MB"
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return feature_not_implemented_yet()
    else:
        return unsupported_exception()


def gpu_used_memory():
    if sys.platform == 'win32':
        for gpu in gpus:
            return f"{gpu.memoryUsed}MB"
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return feature_not_implemented_yet()
    else:
        return unsupported_exception()


def gpu_total_memory():
    if sys.platform == 'win32':
        for gpu in gpus:
            return f"{gpu.memoryTotal}MB"
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return feature_not_implemented_yet()
    else:
        return unsupported_exception()


def gpu_temperature():
    if sys.platform == 'win32':
        gpu = GPUtil.getGPUs()[0]
        return str(gpu.temperature) + 'C'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return feature_not_implemented_yet()
    else:
        return unsupported_exception()



#! /usr/bin/env python
#  -*- coding: utf-8 -*-

# This file is part of windows_tools module

"""
List of installed software from Uninstall registry keys, 32 and 64 bits

Versioning semantics:
    Major version: backward compatibility breaking changes
    Minor version: New functionality
    Patch version: Backwards compatible bug fixes

"""

__intname__ = 'windows_tools.installed_software'
__author__ = 'Orsiris de Jong'
__copyright__ = 'Copyright (C) 2019-2020 Orsiris de Jong'
__description__ = 'List of installed software from Uninstall registry keys, 32 and 64 bits'
__licence__ = 'BSD 3 Clause'
__version__ = '0.5.4'
__build__ = '2021012601'

import windows_tools.registry


def get_installed_software() -> list:
    """
    Returns installed windows software found in registry

    :return: (list) List of dictionnaries containing name, version and publisher of software
    """
    software = []
    try:
        software = software + windows_tools.registry.get_values(hive=windows_tools.registry.HKEY_LOCAL_MACHINE,
                                                                key=r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall',
                                                                names=['DisplayName', 'Publisher', 'DisplayVersion'],
                                                                arch=windows_tools.registry.KEY_WOW64_32KEY | windows_tools.registry.KEY_WOW64_64KEY,
                                                                combine=True)
    except FileNotFoundError:
        pass

    try:
        software = software + windows_tools.registry.get_values(hive=windows_tools.registry.HKEY_CURRENT_USER,
                                                                key=r'SOFTWARE\Microsoft\Windows\CurrentVersion\Uninstall',
                                                                names=['DisplayName', 'Publisher', 'DisplayVersion'],
                                                                arch=windows_tools.registry.KEY_WOW64_32KEY | windows_tools.registry.KEY_WOW64_64KEY,
                                                                combine=True)
    except FileNotFoundError:
        pass

    # map dictionnary keys to actual used values
    for entry in software:
        try:
            entry['name'] = entry.pop('DisplayName')
        except KeyError:
            # Since we don't have a name, we cannot create an entry
            entry['name'] = ''
        try:
            entry['version'] = entry.pop('DisplayVersion')
        except KeyError:
            entry['version'] = ''
        try:
            entry['publisher'] = entry.pop('Publisher')
        except KeyError:
            entry['publisher'] = ''

    # Filter empty results
    return [soft for soft in software if soft != {'name': '', 'version': '', 'publisher': ''}]



import os.path
import sys
import psutil
import wmi
import cpuinfo
from core.exception import *


def cpu_model():
    if sys.platform == 'win32' or 'linux':
        return cpuinfo.get_cpu_info()['brand_raw']
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def cpu_clockspeed():
    if sys.platform == 'win32' or 'linux':
        return cpuinfo.get_cpu_info()['hz_actual_friendly']
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def cpu_architecture():
    if sys.platform == 'win32' or 'linux':
        return cpuinfo.get_cpu_info()['arch']
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def cpu_processor_number():
    if sys.platform == 'win32' or 'linux':
        return cpuinfo.get_cpu_info()['count']
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def cpu_usage():
    if sys.platform == 'win32' or 'linux':
        return str(psutil.cpu_percent()) + '%'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def cpu_temperature():
    if sys.platform == 'win32':
        w_temp = wmi.WMI(namespace="root\\wmi")
        return str(round((w_temp.MSAcpi_ThermalZoneTemperature()[0].CurrentTemperature / 10.0) - 273.15)) + 'C'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        temp = '/sys/devices/virtual/thermal/thermal_zone0/temp'
        if os.path.exists(temp):
            celsius = int(open(temp).read().strip()) / 1000
            return celsius
        else:
            return no_linux_temp_driver()
    else:
        return unsupported_exception()


def cpu_vendor_id():
    if sys.platform == 'win32' or 'linux':
        return cpuinfo.get_cpu_info()['vendor_id_raw']
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2012-2016 Matt Martz
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import warnings

DEPRECATED_MSG = ('The file speedtest_cli.py has been deprecated in favor of '
                  'speedtest.py\nand is available for download at:\n\n'
                  'https://raw.githubusercontent.com/sivel/speedtest-cli/'
                  'master/speedtest.py')


if __name__ == '__main__':
    raise SystemExit(DEPRECATED_MSG)
else:
    try:
        from speedtest import *
    except ImportError:
        raise SystemExit(DEPRECATED_MSG)
    else:
        warnings.warn(DEPRECATED_MSG, UserWarning)



import os
import subprocess
import uuid
import sys
from core.exception import *


def motherboard_model():
    if sys.platform == 'win32':
        model = subprocess.check_output('wmic baseboard get product').decode().split('\n')[1].strip()
        return model
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return os.system('cat /sys/devices/virtual/dmi/id/board_name')
    else:
        return unsupported_exception()


def motherboard_manufacturer():
    if sys.platform == 'win32':
        manufacturer = subprocess.check_output('wmic baseboard get Manufacturer').decode().split('\n')[1].strip()
        return manufacturer
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return os.system('cat /sys/devices/virtual/dmi/id/board_vendor')
    else:
        return unsupported_exception()


def motherboard_serial_number():
    if sys.platform == 'win32':
        serial_id = subprocess.check_output('wmic csproduct get uuid').decode().split('\n')[1].strip()
        return serial_id
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return os.system('cat /sys/class/dmi/id/board_serial')
    else:
        return unsupported_exception()


def motherboard_version():
    if sys.platform == 'win32':
        version = subprocess.check_output('wmic baseboard get version').decode().split('\n')[1].strip()
        return version
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return os.system('/sys/class/dmi/id/board_version:')
    else:
        return unsupported_exception()


def motherboard_node():
    if sys.platform == 'win32' or 'linux':
        return uuid.getnode()
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()



'''Edid data types'''

import json

from collections import namedtuple
from typing import List, NamedTuple, Optional, Tuple, Union


class Edid(NamedTuple):
    '''Parsed EDID object'''
    manufacturer_id: int
    manufacturer: str
    product_id: int
    year: int
    week: int
    edid_version: str
    type: str
    width: float
    height: float
    gamma: float
    dpms_standby: bool
    dpms_suspend: bool
    dpms_activeoff: bool
    resolutions: List[Tuple[int, int, float]]
    name: Optional[str]
    serial: Union[str, int]

    def __repr__(self) -> str:
        return f"Edid({self.manufacturer} {self.name}, s/n {self.serial})"

    def __str__(self) -> str:
        return json.dumps(self._asdict())

_RAW_EDID = namedtuple(
    '_RAW_EDID',
    (
        'header',
        'manu_id',
        'prod_id',
        'serial_no',
        'manu_week',
        'manu_year',
        'edid_version',
        'edid_revision',
        'input_type',
        'width',
        'height',
        'gamma',
        'features',
        'color',
        'timings_supported',
        'timings_reserved',
        'timings_edid',
        'timing_1',
        'timing_2',
        'timing_3',
        'timing_4',
        'extension',
        'checksum'
    )
)


_EDID_STRUCT_FORMAT = (
    '>'     # big-endian
    '8s'    # constant header (8 bytes)
    'H'     # manufacturer id (2 bytes)
    'H'     # product id (2 bytes)
    'I'     # serial number (4 bytes)
    'B'     # manufactoring week (1 byte)
    'B'     # manufactoring year (1 byte)
    'B'     # edid version (1 byte)
    'B'     # edid revision (1 byte)
    'B'     # video input type (1 byte)
    'B'     # horizontal size in cm (1 byte)
    'B'     # vertical size in cm (1 byte)
    'B'     # display gamma (1 byte)
    'B'     # supported features (1 byte)
    '10s'   # color characteristics (10 bytes)
    'H'     # supported timings (2 bytes)
    'B'     # reserved timing (1 byte)
    '16s'   # EDID supported timings (16 bytes)
    '18s'   # detailed timing block 1 (18 bytes)
    '18s'   # detailed timing block 2 (18 bytes)
    '18s'   # detailed timing block 3 (18 bytes)
    '18s'   # detailed timing block 4 (18 bytes)
    'B'     # extension flag (1 byte)
    'B'     # checksum (1 byte)
)

_TIMINGS = {
    0: (1280, 1024, 75.),
    1: (1024,  768, 75.),
    2: (1024,  768, 72.),
    3: (1024,  768, 60.),
    4: (1024,  768, 87.),
    5: ( 832,  624, 75.),
    6: ( 800,  600, 75.),
    7: ( 800,  600, 70.),
    8: ( 800,  600, 60.),
    9: ( 800,  600, 56.),
    10:( 640,  480, 75.),
    11:( 640,  480, 72.),
    12:( 640,  480, 67.),
    13:( 640,  480, 60.),
    14:( 720,  400, 88.),
    15:( 720,  400, 70.),
}

_ASPECT_RATIOS = {
    0b00: (16, 10),
    0b01: ( 4,  3),
    0b10: ( 5,  4),
    0b11: (16,  9),
}



import sys


_enabled = sys.stdout.isatty()
_styles = {
    'bold': 1,
    'dim': 2,
    'italic': 3,
    'underline': 4,
    'inverse': 7,
    'hidden': 8,
    'strikethrough': 9,
    'black': 30,
    'red': 31,
    'green': 32,
    'yellow': 33,
    'blue': 34,
    'magenta': 35,
    'cyan': 36,
    'white': 37,
    'light_black': 90,
    'light_red': 91,
    'light_green': 92,
    'light_yellow': 93,
    'light_blue': 94,
    'light_magenta': 95,
    'light_cyan': 96,
    'light_white': 97,
    'on_black': 40,
    'on_red': 41,
    'on_green': 42,
    'on_yellow': 43,
    'on_blue': 44,
    'on_magenta': 45,
    'on_cyan': 46,
    'on_white': 47,
    'on_light_black': 100,
    'on_light_red': 101,
    'on_light_green': 102,
    'on_light_yellow': 103,
    'on_light_blue': 104,
    'on_light_magenta': 105,
    'on_light_cyan': 106,
    'on_light_white': 107
}


class StyleBuilder:
    def __init__(self, style_list, is_root=False):
        self._style_list = style_list
        self._is_root = is_root

    def __call__(self, *objects, **kwargs):
        sep = kwargs.get('sep', ' ')
        if type(sep) is not str:
            raise TypeError('sep must be None or a string, not %s' % type(sep).__name__)

        string = sep.join(str(obj) for obj in objects)
        if _enabled and self._style_list:
            return '\033[%sm%s\033[0m' % (';'.join([str(val) for val in self._style_list]), string)
        return string

    def __getattr__(self, attr):
        if attr in _styles:
            if self._is_root:
                new_style_list = self._style_list[:]
                new_style_list.append(_styles[attr])
                return StyleBuilder(new_style_list)
            self._style_list.append(_styles[attr])
            return self
        raise AttributeError('%r object has no attribute %r' % (self.__class__.__name__, attr))



import os
import sys
import psutil
from core.exception import *


def drive_list():
    if sys.platform == 'linux':
        disk_info = []
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            disk_info.append({
                'device': part.device,
            })
        return disk_info
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif not sys.platform == 'linux':
        return not_for_linux()
    else:
        return unsupported_exception()


def get_total_space(drive_letter):
    if sys.platform == 'win32':
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            usage = psutil.disk_usage(part.mountpoint)
            if part.device.startswith(drive_letter):
                return str(round(usage.total / 1024 ** 3)) + 'GB'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        obj_disk = psutil.disk_usage('/')
        return str(round(obj_disk.total / (1024.0 ** 3))) + 'GB'
    else:
        return unsupported_exception()


def get_used_space(drive_letter):
    if sys.platform == 'win32':
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            usage = psutil.disk_usage(part.mountpoint)
            if part.device.startswith(drive_letter):
                return str(round(usage.used / 1024 ** 3)) + 'GB'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        obj_disk = psutil.disk_usage('/')
        return str(round(obj_disk.used / (1024.0 ** 3))) + 'GB'
    else:
        return unsupported_exception()


def get_free_space(drive_letter):
    if sys.platform == 'win32':
        return str(int(get_total_space(drive_letter).replace('GB', '')) - int(get_used_space(drive_letter).replace('GB', ''))) + 'GB'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        obj_disk = psutil.disk_usage('/')
        return str(round(obj_disk.free / (1024.0 ** 3))) + 'GB'
    else:
        return unsupported_exception()


def get_used_space_percent(drive_letter):
    if sys.platform == 'win32':
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            usage = psutil.disk_usage(part.mountpoint)
            if part.device.startswith(drive_letter):
                return str(usage.percent) + '%'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        obj_disk = psutil.disk_usage('/')
        return str(round(obj_disk.percent)) + '%'
    else:
        return unsupported_exception()


def get_drive_fstype(drive_letter):
    if sys.platform == 'win32':
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            if part.device.startswith(drive_letter):
                return part.fstype
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return not_for_linux()
    else:
        return unsupported_exception()


def get_drive_mountpoint(drive_letter):
    if sys.platform == 'win32':
        for part in psutil.disk_partitions(all=False):
            if os.name == 'nt':
                if 'cdrom' in part.opts or part.fstype == '':
                    continue
            if part.device.startswith(drive_letter):
                return part.mountpoint
    elif sys.platform == 'darwin':
        return unsupported_exception()
    elif sys.platform == 'linux':
        return not_for_linux()
    else:
        return unsupported_exception()



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# The colorzero color library
#
# Copyright (c) 2018 Dave Jones <dave@waveform.org.uk>
#
# SPDX-License-Identifier: BSD-3-Clause

"Defines various easing functions for :meth:`Color.gradient`."


def linear(steps):
    "Linear easing function; yields *steps* values between 0.0 and 1.0"
    for t in range(steps):
        yield t / (steps - 1)


def ease_in(steps):
    "Quadratic ease-in function; yields *steps* values between 0.0 and 1.0"
    for t in linear(steps):
        yield t ** 2


def ease_out(steps):
    "Quadratic ease-out function; yields *steps* values between 0.0 and 1.0"
    for t in linear(steps):
        yield t * (2 - t)


def ease_in_out(steps):
    "Quadratic ease-in-out function; yields *steps* values between 0.0 and 1.0"
    for t in linear(steps):
        yield 2 * t * t if t < 0.5 else (4 - 2 * t) * t - 1



def unsupported_exception():
    return "\033[91mError: PySil library does not support your operating system.\033[0m"


def unsupported_func():
    return "\033[91mError: Currently function your trying to use is unavailable for your operating system.\033[0m"


def battery_plugged_error():
    return "\033[91mError: Cannot detect remaining battery time because your pc is plugged in.\033[0m"


def no_battery_left_error():
    return "\033[91mError: Cannot detect remaining batter, because its about to run out.\033[0m"


def feature_not_implemented_yet():
    return "\033[91mError: Feature not implemented yet - should be available in next version.\033[0m"


def no_linux_temp_driver():
    return "\033[91mError: Your Linux does not have driver required to get cpu temp.\033[0m"


def not_linux():
    return "\033[91mError: To run this function your os must be Linux.\033[0m"


def not_for_linux():
    return "\033[91mError: You can't use this function in Linux.\033[0m"



# vim: set et sw=4 sts=4 fileencoding=utf-8:
#
# The colorzero color library
#
# Copyright (c) 2016-2021 Dave Jones <dave@waveform.org.uk>
#
# SPDX-License-Identifier: BSD-3-Clause

"Defines the main :class:`Color` class of the package."

import re

from . import conversions as cv, types, attr, deltae, tables, easings

# Lots of the methods below use single character parameter names (r for red, y
# for luma, etc.); this is is normal and in keeping with most of the referenced
# sources
# pylint: disable=invalid-name


class Color(types.RGB):
    """
    The Color class is a tuple which represents a color as linear red, green,
    and blue components.

    The class has a flexible constructor which allows you to create an instance
    from any built-in color system. There are also explicit constructors for
    every known system that can convert (directly or indirectly) to linear RGB.
    For example, an instance of :class:`Color` can be constructed in any of the
    following ways::

        >>> Color('#f00')
        <Color html='#ff0000' rgb=(1, 0, 0)>
        >>> Color('green')
        <Color html='#008000' rgb=(0.0, 0.501961, 0.0)>
        >>> Color(0, 0, 1)
        <Color html='#0000ff' rgb=(0, 0, 1)>
        >>> Color(h=0, s=1, v=0.5)
        <Color html='#800000' rgb=(0.5, 0, 0)>
        >>> Color(y=0.4, u=-0.05, v=0.615)
        <Color html='#ff104c' rgb=(1, 0.0626644, 0.298394)>

    The specific forms that the default constructor will accept are enumerated
    below:

    .. tabularcolumns:: |p{60mm}|p{70mm}|

    +------------------------------+------------------------------------------+
    | Style                        | Description                              |
    +==============================+==========================================+
    | Single scalar parameter      | Equivalent to calling                    |
    |                              | :meth:`Color.from_string`, or            |
    |                              | :meth:`Color.from_rgb24`.                |
    +------------------------------+------------------------------------------+
    | Three positional parameters  | Equivalent to calling                    |
    | or a 3-tuple with no field   | :meth:`Color.from_rgb` if all three      |
    | names                        | parameters are between 0.0 and 1.0, or   |
    |                              | :meth:`Color.from_rgb_bytes` otherwise.  |
    +------------------------------+                                          |
    | Three named parameters, or a |                                          |
    | 3-tuple with fields          |                                          |
    | "r", "g", "b"                |                                          |
    +------------------------------+                                          |
    | Three named parameters, or a |                                          |
    | 3-tuple with fields          |                                          |
    | "red", "green", "blue"       |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_yuv` if "y" is between |
    | "y", "u", "v"                | 0.0 and 1.0, "u" is between -0.436 and   |
    |                              | 0.436, and "v" is between -0.615 and     |
    |                              | 0.615, or :meth:`Color.from_yuv_bytes`   |
    |                              | otherwise.                               |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_yiq`.                  |
    | "y", "i", "q"                |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_hls`.                  |
    | "h", "l", "s"                |                                          |
    +------------------------------+                                          |
    | Three named parameters, or a |                                          |
    | 3-tuple with fields          |                                          |
    | "hue", "lightness",          |                                          |
    | "saturation"                 |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_hsv`                   |
    | "h", "s", "v"                |                                          |
    +------------------------------+                                          |
    | Three named parameters, or a |                                          |
    | 3-tuple with fields          |                                          |
    | "hue", "saturation", "value" |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_xyz`                   |
    | "x", "y", "z"                |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_lab`                   |
    | "l", "a", "b"                |                                          |
    +------------------------------+------------------------------------------+
    | Three named parameters, or a | Equivalent to calling                    |
    | 3-tuple with fields          | :meth:`Color.from_luv`                   |
    | "l", "u", "v"                |                                          |
    +------------------------------+------------------------------------------+

    If the constructor parameters do not conform to any of the variants in the
    table above, a :exc:`ValueError` will be raised.

    Internally, the color is *always* represented as 3 :class:`float` values
    corresponding to the red, green, and blue components of the color. These
    values take a value from 0.0 to 1.0 (least to full intensity). The class
    provides several attributes which can be used to convert one color system
    into another::

        >>> Color('#f00').hls
        HLS(h=0.0, l=0.5, s=1.0)
        >>> Color.from_string('green').hue
        Hue(deg=120.0)
        >>> Color.from_rgb_bytes(0, 0, 255).yuv
        YUV(y=0.114, u=0.436, v=-0.10001426533523537)

    As :class:`Color` derives from tuple, instances are immutable. While this
    provides the advantage that they can be used in a :class:`set` or as keys
    of a :class:`dict`, it does mean that colors themselves cannot be
    *directly* manipulated (e.g. by setting the red component).

    However, several auxilliary classes in the module provide the ability to
    perform simple transformations of colors via operators which produce a new
    :class:`Color` instance. For example, you can add, subtract, and multiply
    colors directly::

        >>> Color('red') + Color('blue')
        <Color html='#ff00ff' rgb=(1, 0, 1)>
        >>> Color('magenta') - Color('red')
        <Color html='#0000ff' rgb=(0, 0, 1)>

    Values are clipped to ensure the resulting color is still valid::

        >>> Color('#ff00ff') + Color('#ff0000')
        <Color html='#ff00ff' rgb=(1, 0, 1)>

    You can wrap numbers in constructors like :class:`Red` (or obtain elements
    of existing colors), then add, subtract, or multiply them with a
    :class:`Color`::

        >>> Color('red') - Red(0.5)
        <Color html='#800000' rgb=(0.5, 0, 0)>
        >>> Color('green') + Color('grey').red
        <Color html='#808000' rgb=(0.501961, 0.501961, 0)>

    You can even manipulate non-primary attributes like hue, saturation, and
    lightness with standard addition, subtraction or multiplication operators::

        >>> Color.from_hls(0.5, 0.5, 1.0)
        <Color html='#00ffff' rgb=(0, 1, 1)>
        >>> Color.from_hls(0.5, 0.5, 1.0) * Lightness(0.8)
        <Color html='#00cccc' rgb=(0, 0.8, 0.8)>
        >>> (Color.from_hls(0.5, 0.5, 1.0) * Lightness(0.8)).hls
        HLS(h=0.5, l=0.4, s=1.0)

    In the last example above, a :class:`Color` instance is constructed from
    HLS (hue, lightness, saturation) values with a lightness of 0.5. This is
    multiplied by a :class:`Lightness` a value of 0.8 which constructs a new
    :class:`Color` with the same hue and saturation, but a lightness of 0.4
    (0.8 × 0.5).

    If an instance is converted to a string (with :func:`str`) it will return a
    string containing the 7-character HTML code for the color (e.g. "#ff0000"
    for red). As can be seen in the examples above, a similar representation is
    included for the output of :func:`repr`. The output of :func:`repr` can
    be customized by assigning values to :attr:`Color.repr_style`.

    .. _RGB: https://en.wikipedia.org/wiki/RGB_color_space
    .. _Y'UV: https://en.wikipedia.org/wiki/YUV
    .. _Y'IQ: https://en.wikipedia.org/wiki/YIQ
    .. _HLS: https://en.wikipedia.org/wiki/HSL_and_HSV
    .. _HSV: https://en.wikipedia.org/wiki/HSL_and_HSV

    .. attribute:: red

        Return the red value as a :class:`Red` instance

    .. attribute:: green

        Return the green value as a :class:`Green` instance

    .. attribute:: blue

        Return the blue value as a :class:`Blue` instance

    .. attribute:: repr_style

        Specifies the style of output returned when using :func:`repr` against
        a :class:`Color` instance. This is an attribute of the class, not of
        instances. For example::

            >>> Color('#f00')
            <Color html='#ff0000' rgb=(1, 0, 0)>
            >>> Color.repr_style = 'html'
            >>> Color('#f00')
            Color('#ff0000')

        The following values are valid:

        * 'default' - The style shown above
        * 'term16m' - Similar to the default style, but instead of the HTML
          style being included, a swatch previewing the color is output. Note
          that the terminal must support `24-bit color ANSI codes`_ for this to
          work.
        * 'term256' - Similar to 'termtrue', but uses the closest color that
          can be found in the standard 256-color xterm palette. Note that the
          terminal must support `8-bit color ANSI codes`_ for this to work.
        * 'html' - Outputs a valid :class:`Color` constructor using the HTML
          style, e.g. ``Color('#ff99bb')``
        * 'rgb' - Outputs a valid :class:`Color` constructor using the floating
          point RGB values, e.g. ``Color(1, 0.25, 0)``
    """
    # pylint: disable=too-many-public-methods

    __slots__ = ()

    repr_style = 'default'

    def __new__(cls, *args, **kwargs):
        def from_rgb(r, g, b):
            "Determine whether bytes or floats are being passed for RGB"
            if 0.0 <= r <= 1.0 and 0.0 <= g <= 1.0 and 0.0 <= b <= 1.0:
                return cls.from_rgb(r, g, b)
            else:
                return cls.from_rgb_bytes(r, g, b)

        def from_yuv(y, u, v):
            "Determine whether bytes or floats are being passed for YUV"
            if (
                    0.0 <= y <= 1.0 and
                    abs(u) <= cv.BT601.Umax and
                    abs(v) <= cv.BT601.Vmax):
                return cls.from_yuv(y, u, v)
            else:
                return cls.from_yuv_bytes(y, u, v)

        if kwargs:
            try:
                # Yes, lambdas are fine here
                # pylint: disable=unnecessary-lambda
                return {
                    frozenset('rgb'): from_rgb,
                    frozenset('yuv'): from_yuv,
                    frozenset('yiq'): cls.from_yiq,
                    frozenset('hls'): cls.from_hls,
                    frozenset('hsv'): cls.from_hsv,
                    frozenset('xyz'): cls.from_xyz,
                    frozenset('lab'): cls.from_lab,
                    frozenset('luv'): cls.from_luv,
                    frozenset('cmy'): cls.from_cmy,
                    frozenset('cmyk'): cls.from_cmyk,
                    frozenset(('red', 'green', 'blue')):
                        lambda red, green, blue:
                        from_rgb(red, green, blue),
                    frozenset(('cyan', 'magenta', 'yellow')):
                        lambda cyan, magenta, yellow:
                        cls.from_cmy(cyan, magenta, yellow),
                    frozenset(('cyan', 'magenta', 'yellow', 'black')):
                        lambda cyan, magenta, yellow, black:
                        cls.from_cmyk(cyan, magenta, yellow, black),
                    frozenset(('hue', 'lightness', 'saturation')):
                        lambda hue, lightness, saturation:
                        cls.from_hls(hue, lightness, saturation),
                    frozenset(('hue', 'saturation', 'value')):
                        lambda hue, saturation, value:
                        cls.from_hsv(hue, saturation, value),
                    }[frozenset(kwargs.keys())](**kwargs)
            except KeyError:
                pass
        else:
            if len(args) == 1:
                if isinstance(args[0], bytes):
                    spec = args[0].decode('ascii')
                else:
                    spec = args[0]
                if isinstance(spec, str):
                    return cls.from_string(spec)
                elif isinstance(spec, tuple):
                    try:
                        return cls(**spec._asdict())
                    except AttributeError:
                        if len(spec) == 3:
                            return from_rgb(*spec)
                elif isinstance(spec, int):
                    return cls.from_rgb24(spec)
            elif len(args) == 3:
                r, g, b = args
                return from_rgb(r, g, b)
        raise ValueError('Unable to construct Color from provided arguments')

    @classmethod
    def from_string(cls, s):
        """
        Construct a :class:`Color` from a 4 or 7 character CSS-like
        representation (e.g. "#f00" or "#ff0000" for red), or from one of the
        named colors (e.g. "green" or "wheat") from the `CSS standard`_. Any
        other string format will result in a :exc:`ValueError`.

        .. _CSS standard: http://www.w3.org/TR/css3-color/#svg-color
        """
        if s[:1] != '#':
            s = cv.name_to_html(s)
        return cls.from_rgb_bytes(*cv.html_to_rgb_bytes(s))

    @classmethod
    def from_rgb(cls, r, g, b):
        """
        Construct a :class:`Color` from three linear `RGB`_ float values
        between 0.0 and 1.0.
        """
        return super().__new__(cls,
                               cv.clamp_float(r),
                               cv.clamp_float(g),
                               cv.clamp_float(b))

    @classmethod
    def from_rgb24(cls, n):
        """
        Construct a :class:`Color` from an unsigned 24-bit integer number
        of the form 0x00BBGGRR.
        """
        return cls.from_rgb_bytes(*cv.rgb24_to_rgb_bytes(n))

    @classmethod
    def from_rgb565(cls, n):
        """
        Construct a :class:`Color` from an unsigned 16-bit integer number
        in RGB565 format.
        """
        return cls.from_rgb(*cv.rgb565_to_rgb(n))

    @classmethod
    def from_rgb_bytes(cls, r, g, b):
        """
        Construct a :class:`Color` from three `RGB`_ byte values between 0 and
        255.

        .. _RGB: https://en.wikipedia.org/wiki/RGB_color_space
        """
        return cls.from_rgb(*cv.rgb_bytes_to_rgb(r, g, b))

    @classmethod
    def from_yuv(cls, y, u, v):
        """
        Construct a :class:`Color` from three `Y'UV`_ float values. The Y value
        may be between 0.0 and 1.0. U may be between -0.436 and 0.436, while
        V may be between -0.615 and 0.615.

        .. _Y'UV: https://en.wikipedia.org/wiki/YUV
        """
        return cls.from_rgb(*cv.yuv_to_rgb(y, u, v))

    @classmethod
    def from_yuv_bytes(cls, y, u, v):
        """
        Construct a :class:`Color` from three `Y'UV`_ byte values between 0 and
        255. The U and V values are biased by 128 to prevent negative values as
        is typical in video applications. The Y value is biased by 16 for the
        same purpose.

        .. _Y'UV: https://en.wikipedia.org/wiki/YUV
        """
        return cls.from_rgb_bytes(*cv.yuv_bytes_to_rgb_bytes(y, u, v))

    @classmethod
    def from_yiq(cls, y, i, q):
        """
        Construct a :class:`Color` from three `Y'IQ`_ float values. Y' can be
        between 0.0 and 1.0, while I and Q can be between -1.0 and 1.0.

        .. _Y'IQ: https://en.wikipedia.org/wiki/YIQ
        """
        return cls.from_rgb(*cv.yiq_to_rgb(y, i, q))

    @classmethod
    def from_hls(cls, h, l, s):
        """
        Construct a :class:`Color` from `HLS`_ (hue, lightness, saturation)
        floats between 0.0 and 1.0.

        .. _HLS: https://en.wikipedia.org/wiki/HSL_and_HSV
        """
        return cls.from_rgb(*cv.hls_to_rgb(h, l, s))

    @classmethod
    def from_hsv(cls, h, s, v):
        """
        Construct a :class:`Color` from `HSV`_ (hue, saturation, value) floats
        between 0.0 and 1.0.

        .. _HSV: https://en.wikipedia.org/wiki/HSL_and_HSV
        """
        return cls.from_rgb(*cv.hsv_to_rgb(h, s, v))

    @classmethod
    def from_cmy(cls, c, m, y):
        """
        Construct a :class:`Color` from `CMY`_ (cyan, magenta, yellow) floats
        between 0.0 and 1.0.

        .. note::

            This conversion uses the basic subtractive method which is not
            accurate for color reproduction on print devices. See the `Color
            FAQ`_ for more information.

        .. _Color FAQ: http://poynton.ca/notes/colour_and_gamma/ColorFAQ.html#RTFToC24
        .. _CMY: https://en.wikipedia.org/wiki/CMYK_color_model
        """
        return cls.from_rgb(*cv.cmy_to_rgb(c, m, y))

    @classmethod
    def from_cmyk(cls, c, m, y, k):
        """
        Construct a :class:`Color` from `CMYK`_ (cyan, magenta, yellow, black)
        floats between 0.0 and 1.0.

        .. note::

            This conversion uses the basic subtractive method which is not
            accurate for color reproduction on print devices. See the `Color
            FAQ`_ for more information.

        .. _Color FAQ: http://poynton.ca/notes/colour_and_gamma/ColorFAQ.html#RTFToC24
        .. _CMYK: https://en.wikipedia.org/wiki/CMYK_color_model
        """
        return cls.from_cmy(*cv.cmyk_to_cmy(c, m, y, k))

    @classmethod
    def from_xyz(cls, x, y, z):
        """
        Construct a :class:`Color` from (X, Y, Z) float values representing
        a color in the `CIE 1931 color space`_. The conversion assumes the
        sRGB working space with reference white D65.

        .. _CIE 1931 color space: https://en.wikipedia.org/wiki/CIE_1931_color_space
        """
        return cls.from_rgb(*cv.xyz_to_rgb(x, y, z))

    @classmethod
    def from_lab(cls, l, a, b):
        """
        Construct a :class:`Color` from (L*, a*, b*) float values representing
        a color in the `CIE Lab color space`_. The conversion assumes the
        sRGB working space with reference white D65.

        .. _CIE Lab color space: https://en.wikipedia.org/wiki/Lab_color_space
        """
        return cls.from_xyz(*cv.lab_to_xyz(l, a, b))

    @classmethod
    def from_luv(cls, l, u, v):
        """
        Construct a :class:`Color` from (L*, u*, v*) float values representing
        a color in the `CIE Luv color space`_. The conversion assumes the sRGB
        working space with reference white D65.

        .. _CIE Luv color space: https://en.wikipedia.org/wiki/CIELUV
        """
        return cls.from_xyz(*cv.luv_to_xyz(l, u, v))

    def __add__(self, other):
        if isinstance(other, types.RGB):
            return Color.from_rgb(self.r + other.r,
                                  self.g + other.g,
                                  self.b + other.b)
        elif isinstance(other, (attr.Red, attr.Green, attr.Blue)):
            r, g, b = self
            return Color.from_rgb(
                r + other if isinstance(other, attr.Red) else r,
                g + other if isinstance(other, attr.Green) else g,
                b + other if isinstance(other, attr.Blue) else b,
            )
        elif isinstance(other, (attr.Hue, attr.Lightness, attr.Saturation)):
            h, l, s = self.hls
            return Color.from_hls(
                h + other if isinstance(other, attr.Hue) else h,
                l + other if isinstance(other, attr.Lightness) else l,
                s + other if isinstance(other, attr.Saturation) else s,
            )
        elif isinstance(other, attr.Luma):
            y, u, v = self.yuv
            return Color.from_yuv(y + other, u, v)
        return NotImplemented

    def __radd__(self, other):
        # Addition is commutative
        if isinstance(other, (types.RGB,
                              attr.Red, attr.Green, attr.Blue,
                              attr.Hue, attr.Lightness, attr.Saturation,
                              attr.Luma)):
            return self.__add__(other)
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, types.RGB):
            return Color.from_rgb(self.r - other.r,
                                  self.g - other.g,
                                  self.b - other.b)
        elif isinstance(other, (attr.Red, attr.Green, attr.Blue)):
            r, g, b = self.rgb
            return Color.from_rgb(
                r - other if isinstance(other, attr.Red) else r,
                g - other if isinstance(other, attr.Green) else g,
                b - other if isinstance(other, attr.Blue) else b,
            )
        elif isinstance(other, (attr.Hue, attr.Lightness, attr.Saturation)):
            h, l, s = self.hls
            return Color.from_hls(
                h - other if isinstance(other, attr.Hue) else h,
                l - other if isinstance(other, attr.Lightness) else l,
                s - other if isinstance(other, attr.Saturation) else s,
            )
        elif isinstance(other, attr.Luma):
            y, u, v = self.yuv
            return Color.from_yuv(y - other, u, v)
        return NotImplemented

    def __rsub__(self, other):
        if isinstance(other, (attr.Red, attr.Green, attr.Blue)):
            r, g, b = self.rgb
            return Color.from_rgb(
                other - r if isinstance(other, attr.Red) else 0.0,
                other - g if isinstance(other, attr.Green) else 0.0,
                other - b if isinstance(other, attr.Blue) else 0.0,
            )
        return NotImplemented

    def __mul__(self, other):
        if isinstance(other, types.RGB):
            return Color.from_rgb(self.r * other.r,
                                  self.g * other.g,
                                  self.b * other.b)
        elif isinstance(other, (attr.Red, attr.Green, attr.Blue)):
            r, g, b = self
            return Color.from_rgb(
                r * other if isinstance(other, attr.Red) else r,
                g * other if isinstance(other, attr.Green) else g,
                b * other if isinstance(other, attr.Blue) else b,
            )
        elif isinstance(other, (attr.Hue, attr.Lightness, attr.Saturation)):
            h, l, s = self.hls
            return Color.from_hls(
                h * other if isinstance(other, attr.Hue) else h,
                l * other if isinstance(other, attr.Lightness) else l,
                s * other if isinstance(other, attr.Saturation) else s,
            )
        elif isinstance(other, attr.Luma):
            y, u, v = self.yuv
            return Color.from_yuv(y * other, u, v)
        return NotImplemented

    def __rmul__(self, other):
        # Multiplication is commutative
        if isinstance(other, (types.RGB,
                              attr.Red, attr.Green, attr.Blue,
                              attr.Hue, attr.Lightness, attr.Saturation,
                              attr.Luma)):
            return self.__mul__(other)
        return NotImplemented

    _format_re = re.compile(
        r'^('
        r'(?P<html>html)|'
        r'(?P<css>css(?P<cssfmt>rgb|hsl)?)|'
        r'(?P<back>[fb])?(?P<term>0|8|256|16[mM])?'
        r')$')
    def __format__(self, format_spec):
        m = Color._format_re.match(format_spec)
        if not m:
            raise ValueError(
                'Invalid format {:r} for Color'.format(format_spec))
        if m.group('html'):
            return self.html
        elif m.group('css'):
            return self._format_css(m.group('cssfmt'))
        else:
            return self._format_term(m.group('back'), m.group('term'))

    def _format_css(self, cssfmt):
        return {
            None:  lambda self: 'rgb({0:d}, {1:d}, {2:d})'.format(
                *self.rgb_bytes),
            'rgb': lambda self: 'rgb({0:d}, {1:d}, {2:d})'.format(
                *self.rgb_bytes),
            'hsl': lambda self: 'hsl({0:g}deg, {1:g}%, {2:g}%)'.format(
                self.hls.hue.deg, self.hls.saturation * 100,
                self.hls.lightness * 100),
        }[cssfmt](self)

    def _format_term(self, back, term):
        if term == '0':
            args = ({
                None: 0,
                'f':  39,
                'b':  49,
            }[back],)
        elif term in (None, '8'):
            table = tables.DOS_COLORS
            if back == 'b':
                code = 40
                table = {
                    k: (bold, index)
                    for k, (bold, index) in table.items()
                    if not bold
                }
            else:
                code = 30
            try:
                bold, index = table[self.rgb_bytes]
            except KeyError:
                bold, index = sorted(
                    (self.difference(Color.from_rgb_bytes(*color)), bold, index)
                    for color, (bold, index) in table.items()
                )[0][1:]
            args = (1,) if bold else ()
            args += (code + index,)
        elif term == '256':
            code = 48 if back == 'b' else 38
            try:
                index = tables.XTERM_COLORS[self.rgb_bytes]
            except KeyError:
                index = sorted(
                    (self.difference(Color.from_rgb_bytes(*color)), index)
                    for color, index in tables.XTERM_COLORS.items()
                )[0][1]
            args = (48 if back == 'b' else 38, 5, index)
        elif term.lower() == '16m':
            args = (48 if back == 'b' else 38, 2) + self.rgb_bytes
        else:
            assert False  # pragma: no cover
        return '\x1b[' + ';'.join(str(i) for i in args) + 'm'

    def __str__(self):
        return self.html

    def __repr__(self):
        try:
            return {
                'default': '<Color html={self.html!r} '
                           'rgb=({self.r:g}, {self.g:g}, {self.b:g})>',
                'term16m': '<Color {self:16m}###{self:0} '
                           'rgb=({self.r:g}, {self.g:g}, {self.b:g})>',
                'term256': '<Color {self:256}###{self:0} '
                           'rgb=({self.r:g}, {self.g:g}, {self.b:g})>',
                'html':    'Color({self.html!r})',
                'rgb':     'Color({self.r:g}, {self.g:g}, {self.b:g})',
            }[Color.repr_style].format(self=self)
        except KeyError:
            raise ValueError(
                'invalid repr_style value: {}'.format(Color.repr_style))

    @property
    def html(self):
        """
        Returns the color as a string in HTML #RRGGBB format.
        """
        return cv.rgb_bytes_to_html(*self.rgb_bytes)

    @property
    def rgb(self):
        """
        Return a simple 3-tuple of (r, g, b) float values in the range 0.0 <= n
        <= 1.0.

        .. note::

            The :class:`Color` class can already be treated as such a 3-tuple
            but for the cases where you want a straight
            :func:`~collections.namedtuple` this property is available.
        """
        return types.RGB(*self)

    @property
    def rgb565(self):
        """
        Returns an unsigned 16-bit integer number representing the color in
        the RGB565 encoding.
        """
        return cv.rgb_to_rgb565(*self)

    @property
    def rgb_bytes(self):
        """
        Returns a 3-tuple of (red, green, blue) byte values.
        """
        return cv.rgb_to_rgb_bytes(*self)

    @property
    def yuv(self):
        """
        Returns a 3-tuple of (y, u, v) float values; Y values can be between
        0.0 and 1.0, U values are between -0.436 and 0.436, and V values are
        between -0.615 and 0.615.
        """
        return cv.rgb_to_yuv(*self)

    @property
    def yuv_bytes(self):
        """
        Returns a 3-tuple of (y, u, v) byte values. Y values are biased by 16
        in the result to prevent negatives. U and V values are biased by 128
        for the same purpose.
        """
        return cv.rgb_bytes_to_yuv_bytes(*self.rgb_bytes)

    @property
    def yiq(self):
        """
        Returns a 3-tuple of (y, i, q) float values; y values can be between
        0.0 and 1.0, whilst i and q values can be between -1.0 and 1.0.
        """
        return cv.rgb_to_yiq(*self)

    @property
    def xyz(self):
        """
        Returns a 3-tuple of (X, Y, Z) float values representing the color in
        the `CIE 1931 color space`_. The conversion assumes the sRGB working
        space, with reference white D65.

        .. _CIE 1931 color space: https://en.wikipedia.org/wiki/CIE_1931_color_space
        """
        return cv.rgb_to_xyz(*self)

    @property
    def lab(self):
        """
        Returns a 3-tuple of (L*, a*, b*) float values representing the color
        in the `CIE Lab color space`_ with the `D65 standard illuminant`_.

        .. _CIE Lab color space: https://en.wikipedia.org/wiki/Lab_color_space
        .. _D65 standard illuminant: https://en.wikipedia.org/wiki/Illuminant_D65
        """
        return cv.xyz_to_lab(*self.xyz)

    @property
    def luv(self):
        """
        Returns a 3-tuple of (L*, u*, v*) float values representing the color
        in the `CIE Luv color space`_ with the `D65 standard illuminant`_.

        .. _CIE Luv color space: https://en.wikipedia.org/wiki/CIELUV
        """
        return cv.xyz_to_luv(*self.xyz)

    @property
    def hls(self):
        """
        Returns a 3-tuple of (hue, lightness, saturation) float values (between
        0.0 and 1.0).
        """
        return cv.rgb_to_hls(*self)

    @property
    def hsv(self):
        """
        Returns a 3-tuple of (hue, saturation, value) float values (between 0.0
        and 1.0).
        """
        return cv.rgb_to_hsv(*self)

    @property
    def cmy(self):
        """
        Returns a 3-tuple of (cyan, magenta, yellow) float values (between 0.0
        and 1.0).

        .. note::

            This conversion uses the basic subtractive method which is not
            accurate for color reproduction on print devices. See the `Color
            FAQ`_ for more information.

        .. _Color FAQ: http://poynton.ca/notes/colour_and_gamma/ColorFAQ.html#RTFToC24
        """
        return cv.rgb_to_cmy(*self)

    @property
    def cmyk(self):
        """
        Returns a 4-tuple of (cyan, magenta, yellow, black) float values
        (between 0.0 and 1.0).

        .. note::

            This conversion uses the basic subtractive method which is not
            accurate for color reproduction on print devices. See the `Color
            FAQ`_ for more information.

        .. _Color FAQ: http://poynton.ca/notes/colour_and_gamma/ColorFAQ.html#RTFToC24
        """
        return cv.cmy_to_cmyk(*self.cmy)

    @property
    def hue(self):
        """
        Returns the hue of the color as a :class:`Hue` instance which can be
        used in operations with other :class:`Color` instances.
        """
        return attr.Hue(self.hls[0])

    @property
    def lightness(self):
        """
        Returns the lightness of the color as a :class:`Lightness` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return attr.Lightness(self.hls[1])

    @property
    def saturation(self):
        """
        Returns the saturation of the color as a :class:`Saturation` instance
        which can be used in operations with other :class:`Color` instances.
        """
        return attr.Saturation(self.hls[2])

    @property
    def luma(self):
        """
        Returns the `luma`_ of the color as a :class:`Luma` instance which can
        be used in operations with other :class:`Color` instances.

        .. _luma: https://en.wikipedia.org/wiki/Luma_(video)
        """
        return attr.Luma(self.yuv[0])

    def difference(self, other, method='euclid'):
        """
        Determines the difference between this color and *other* using the
        specified *method*.

        :param Color other:
            The color to compare this color to.

        :param str method:
            The algorithm to use in the comparison. Valid values are:

            * 'euclid' - This is the default method. Calculate the `Euclidian
              distance`_. This is by far the fastest method, but also the least
              accurate in terms of human perception.
            * 'cie1976' - Use the `CIE 1976`_ formula for calculating the
              difference between two colors in CIE Lab space.
            * 'cie1994g' - Use the `CIE 1994`_ formula with the "graphic arts"
              bias for calculating the difference.
            * 'cie1994t' - Use the `CIE 1994`_ forumula with the "textiles"
              bias for calculating the difference.
            * 'ciede2000' - Use the `CIEDE 2000`_ formula for calculating the
              difference.

        :returns:
            A :class:`float` indicating how different the two colors are. Note
            that the Euclidian distance will be significantly different to the
            other calculations; effectively this just measures the distance
            between the two colors by treating them as coordinates in a three
            dimensional Euclidian space. All other methods are means of
            calculating a `Delta E`_ value in which 2.3 is considered a
            `just-noticeable difference`_ (JND).

        For example::

            >>> Color('red').difference(Color('red'))
            0.0
            >>> Color('red').difference(Color('red'), method='cie1976')
            0.0
            >>> Color('red').difference(Color('#900'))
            0.4
            >>> Color('red').difference(Color('#900'), method='cie1976')
            40.17063087142142
            >>> Color('red').difference(Color('#900'), method='ciede2000')
            21.078146289272155
            >>> Color('red').difference(Color('blue'))
            1.4142135623730951
            >>> Color('red').difference(Color('blue'), method='cie1976')
            176.31403908880046

        .. note::

            Instead of using this method, you may wish to simply use the
            various difference functions (:func:`euclid`, :func:`cie1976`,
            etc.) directly.

        .. _Delta E: https://en.wikipedia.org/wiki/Color_difference
        .. _just-noticeable difference: https://en.wikipedia.org/wiki/Just-noticeable_difference
        .. _Euclidian distance: https://en.wikipedia.org/wiki/Euclidean_distance
        .. _CIE 1976: https://en.wikipedia.org/wiki/Color_difference#CIE76
        .. _CIE 1994: https://en.wikipedia.org/wiki/Color_difference#CIE94
        .. _CIEDE 2000: https://en.wikipedia.org/wiki/Color_difference#CIEDE2000
        """
        if isinstance(method, bytes):
            method = method.decode('ascii')
        try:
            fn = getattr(deltae, method)
        except AttributeError:
            raise ValueError('invalid method: {}'.format(method))
        else:
            if method.startswith('cie'):
                return fn(self.lab, other.lab)
            else:
                return fn(self, other)

    def gradient(self, other, steps=10, easing=easings.linear):
        """
        Returns a generator which fades between this color and *other* in the
        specified number of *steps*.

        :param Color other:
            The color that will end the gradient (with the color the method is
            called upon starting the gradient)

        :param int steps:
            The unqiue number of colors to include in the generated gradient.
            Defaults to 10 if unspecified.

        :param callable easing:
            A function which controls the speed of the progression. If
            specified, if must be a function which takes a single parameter,
            the number of *steps*, and yields a sequence of values between 0.0
            (representing the start of the gradient) and 1.0 (representing the
            end). The default is :func:`linear`.

        :return:
            A generator yielding *steps* :class:`Color` instances which fade
            from this color to *other*.

        For example::

            >>> Color.repr_style = 'html'
            >>> print('\\n'.join(
            ... repr(c) for c in
            ... Color('red').gradient(Color('green'))
            ... ))
            Color('#ff0000')
            Color('#e30e00')
            Color('#c61c00')
            Color('#aa2b00')
            Color('#8e3900')
            Color('#714700')
            Color('#555500')
            Color('#396400')
            Color('#1c7200')
            Color('#008000')

        .. versionadded:: 1.1
        """
        if steps < 2:
            raise ValueError('steps must be >= 2')
        # NOTE: Can't simply subtract self from other here, as the result will
        # be clamped and we want the actual result.
        delta = types.RGB(*(
            other_i - self_i
            for self_i, other_i in zip(self, other)
        ))
        for t in easing(steps):
            yield self + types.RGB(*(delta_i * t for delta_i in delta))



from setuptools import setup

classifiers = [
    'Development Status :: 5 - Production/Stable',
    'Operating System :: Microsoft :: Windows :: Windows 10',
    'Programming Language :: Python :: 3'
]

setup(
    name='pysil',
    version='1.1.5',
    description='system information gathering made simple',
    long_description=open('README.txt').read() + '\n\n' + open('CHANGELOG.txt').read(),
    url='https://github.com/Bamboooz/pysil',
    author='Bamboooz',
    author_email='bambusixmc@gmail.com',
    license='MIT',
    classifiers=classifiers,
    keywords='system',
    packages=['pysil', 'core'],
    install_requires=[
        'distro',
        'psutil',
        'windows-tools.antivirus',
        'py-cpuinfo',
        'GPUtil',
        'netifaces',
        'speedtest-cli',
        'screeninfo',
        'xlib',
        'sounddevice',
    ]
)



#! /usr/bin/env python
#  -*- coding: utf-8 -*-

# This file is part of windows_tools module

"""
Windows registry simple API

Versioning semantics:
    Major version: backward compatibility breaking changes
    Minor version: New functionality
    Patch version: Backwards compatible bug fixes

"""

__intname__ = "windows_tools.registry"
__author__ = "Orsiris de Jong"
__copyright__ = "Copyright (C) 2019-2021 Orsiris de Jong"
__description__ = "Windows registry 32 and 64 bits simple API"
__licence__ = "BSD 3 Clause"
__version__ = "1.0.1"
__build__ = "202110501"

from typing import List, NoReturn, Optional, Union

# that import is needed so we get CONSTANTS from winreg (eg HKEY_LOCAL_MACHINE etc) for direct use in module
from winreg import *  # noqa ignore=F405

# The following lines make lint tools happy
from winreg import (
    ConnectRegistry,
    OpenKey,
    EnumKey,
    EnumValue,
    QueryInfoKey,
    QueryValueEx,
    DeleteKey,
)
from winreg import KEY_WOW64_32KEY, KEY_WOW64_64KEY, KEY_READ, KEY_ALL_ACCESS, HKEYType
from windows_tools.misc import windows_ticks_to_date


def get_value(
    hive: int,
    key: str,
    value: Optional[str],
    arch: int = 0,
    last_modified: bool = False,
) -> Union[str, dict]:
    """
    Returns a value from a given registry path

    :param hive: registry hive (windows.registry.HKEY_LOCAL_MACHINE...)
    :param key:  which registry key we're searching for
    :param value: which value we query, may be None if unnamed value is searched
    :param arch: which registry architecture we seek (0 = default, windows.registry.KEY_WOW64_64KEY, windows.registry.KEY_WOW64_32KEY)
                 Giving multiple arches here will return first result
    :return: value
    """

    def _get_value(hive: int, key: str, value: Optional[str], arch: int) -> str:
        try:
            open_reg = ConnectRegistry(None, hive)
            open_key = OpenKey(open_reg, key, 0, KEY_READ | arch)
            if last_modified:
                output = {}
                output["value"], key_type = QueryValueEx(open_key, value)
                timestamp = windows_ticks_to_date(QueryInfoKey(open_key)[2])
                output["last_modified"] = timestamp
            else:
                output, key_type = QueryValueEx(open_key, value)
            # Return the first match
            return output
        except (FileNotFoundError, TypeError, OSError) as exc:
            raise FileNotFoundError(
                "Registry key [%s] with value [%s] not found. %s" % (key, value, exc)
            )

    # 768 = 0 | KEY_WOW64_64KEY | KEY_WOW64_32KEY (where 0 = default)
    if arch == 768:
        for _arch in [KEY_WOW64_64KEY, KEY_WOW64_32KEY]:
            try:
                return _get_value(hive, key, value, _arch)
            except FileNotFoundError:
                pass
        raise FileNotFoundError
    else:
        return _get_value(hive, key, value, arch)


def get_values(
    hive: int,
    key: str,
    names: List[str],
    arch: int = 0,
    combine: bool = False,
    last_modified: bool = False,
) -> list:
    """
    Returns a dictionnary of values in names from registry key

    :param hive: registry hive (windows.registry.HKEY_LOCAL_MACHINE...)
    :param key: which registry key we're searching for
    :param names: which value names we query for
    :param arch: which registry architecture we seek (0 = default, windows.registry.KEY_WOW64_64KEY, windows.registry.KEY_WOW64_32KEY)
    :param combine: shall we combine multiple arch results or return first match
    :return: list of strings
    """

    def _get_values(hive: int, key: str, names: List[str], arch: int) -> list:
        try:
            open_reg = ConnectRegistry(None, hive)
            open_key = OpenKey(open_reg, key, 0, KEY_READ | arch)
            subkey_count, value_count, _ = QueryInfoKey(open_key)

            output = []
            for index in range(subkey_count):
                values = {}
                subkey_name = EnumKey(open_key, index)
                subkey_handle = OpenKey(open_key, subkey_name)
                for name in names:
                    try:
                        if last_modified:
                            values[name] = {}
                            values[name]["value"] = QueryValueEx(subkey_handle, name)[0]
                            timestamp = windows_ticks_to_date(
                                QueryInfoKey(subkey_handle)[2]
                            )
                            values[name]["last_modified"] = timestamp
                        else:
                            values[name] = QueryValueEx(subkey_handle, name)[0]
                    except (FileNotFoundError, TypeError):
                        pass
                output.append(values)
            return output

        except (FileNotFoundError, TypeError, OSError) as exc:
            raise FileNotFoundError("Cannot query registry key [%s]. %s" % (key, exc))

    # 768 = 0 | KEY_WOW64_64KEY | KEY_WOW64_32KEY (where 0 = default)
    if arch == 768:
        result = []
        for _arch in [KEY_WOW64_64KEY, KEY_WOW64_32KEY]:
            try:
                if combine:
                    result = result + (_get_values(hive, key, names, _arch))
                else:
                    return _get_values(hive, key, names, _arch)
            except FileNotFoundError:
                pass
        return result
    else:
        return _get_values(hive, key, names, arch)


OPEN_REGISTRY_HANDLE = None


def get_keys(
    hive: int,
    key: str,
    arch: int = 0,
    recursion_level: int = 1,
    filter_on_names: List[str] = None,
    combine: bool = False,
    last_modified: bool = False,
) -> dict:
    """
    :param hive: registry hive (windows.registry.HKEY_LOCAL_MACHINE...)
    :param key: which registry key we're searching for
    :param arch: which registry architecture we seek (0 = default, windows.registry.KEY_WOW64_64KEY, windows.registry.KEY_WOW64_32KEY)
    :param recursion_level: recursivity level
    :param filter_on_names: list of strings we search, if none given, all value names are returned
    :param combine: shall we combine multiple arch results or return first match
    :return: list of strings
    """

    global OPEN_REGISTRY_HANDLE

    def _get_keys(
        hive: int, key: str, arch: int, recursion_level: int, filter_on_names: List[str]
    ):
        global OPEN_REGISTRY_HANDLE

        try:
            if not OPEN_REGISTRY_HANDLE:
                OPEN_REGISTRY_HANDLE = ConnectRegistry(None, hive)
            open_key = OpenKey(OPEN_REGISTRY_HANDLE, key, 0, KEY_READ | arch)
            subkey_count, value_count, _ = QueryInfoKey(open_key)

            output = {}
            values = []
            for index in range(value_count):
                name, value, type = EnumValue(open_key, index)
                if isinstance(filter_on_names, list) and name not in filter_on_names:
                    pass
                else:
                    if last_modified:
                        last_modified_date = windows_ticks_to_date(
                            QueryInfoKey(open_key)[2]
                        )
                        data = {
                            "name": name,
                            "value": value,
                            "type": type,
                            "last_modified": last_modified_date,
                        }
                    else:
                        data = {"name": name, "value": value, "type": type}
                    values.append(data)
            if not values == []:
                output[""] = values

            if recursion_level > 0:
                for subkey_index in range(subkey_count):
                    try:
                        subkey_name = EnumKey(open_key, subkey_index)
                        sub_values = get_keys(
                            hive=0,
                            key=key + "\\" + subkey_name,
                            arch=arch,
                            recursion_level=recursion_level - 1,
                            filter_on_names=filter_on_names,
                            last_modified=last_modified,
                        )
                        output[subkey_name] = sub_values
                    except FileNotFoundError:
                        pass

            return output

        except (FileNotFoundError, TypeError, OSError) as exc:
            raise FileNotFoundError("Cannot query registry key [%s]. %s" % (key, exc))

    # 768 = 0 | KEY_WOW64_64KEY | KEY_WOW64_32KEY (where 0 = default)
    if arch == 768:
        result = {}
        for _arch in [KEY_WOW64_64KEY, KEY_WOW64_32KEY]:
            try:
                if combine:
                    result.update(
                        _get_keys(hive, key, _arch, recursion_level, filter_on_names)
                    )
                else:
                    return _get_keys(hive, key, _arch, recursion_level, filter_on_names)
            except FileNotFoundError:
                pass
        return result
    else:
        return _get_keys(hive, key, arch, recursion_level, filter_on_names)


def delete_sub_key(root_key: int, current_key: str, arch: int = 0) -> None:
    """

    :param root_key: winreg registry root key constant
    :param current_key:
    :param arch:
    :return:
    """

    def _delete_sub_key(root_key: int, current_key: str, arch: int) -> NoReturn:
        open_key = OpenKey(root_key, current_key, 0, KEY_ALL_ACCESS | arch)
        info_key = QueryInfoKey(open_key)
        for _ in range(0, info_key[0]):
            # NOTE:: This code is to delete the key and all sub_keys.
            # If you just want to walk through them, then
            # you should pass x to EnumKey. sub_key = EnumKey(open_key, x)
            # Deleting the sub_key will change the sub_key count used by EnumKey.
            # We must always pass 0 to EnumKey so we
            # always get back the new first sub_key.
            sub_key = EnumKey(open_key, 0)
            try:
                DeleteKey(open_key, sub_key)
            except OSError:
                _delete_sub_key(root_key, "\\".join([current_key, sub_key]), arch)
                # No extra delete here since each call
                # to delete_sub_key will try to delete itself when its empty.

        DeleteKey(open_key, "")
        open_key.Close()
        return

    # 768 = 0 | KEY_WOW64_64KEY | KEY_WOW64_32KEY (where 0 = default)
    if arch == 768:
        for _arch in [KEY_WOW64_64KEY, KEY_WOW64_32KEY]:
            _delete_sub_key(root_key, current_key, _arch)
    else:
        _delete_sub_key(root_key, current_key, arch)



import ipaddress
import socket
import netifaces
import sys
import speedtest
from core.exception import *


def get_ipv4():
    if sys.platform == 'win32' or 'linux':
        hostname = socket.gethostname()
        ipv4 = socket.gethostbyname(hostname)
        return ipv4
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_ipv6():
    if sys.platform == 'win32' or 'linux':
        alladdr = socket.getaddrinfo(socket.gethostname(), 0)
        ip6 = filter(
            lambda x: x[0] == socket.AF_INET6,  # means its ip6
            alladdr
        )
        return list(ip6)[0][4][0]
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_subnet_mask():
    if sys.platform == 'win32' or 'linux':
        ip_addr = socket.gethostbyname(socket.gethostname())
        netmask = ipaddress.IPv4Network(ip_addr).netmask
        return netmask
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_default_gateway():
    if sys.platform == 'win32' or 'linux':
        gateways = netifaces.gateways()
        defaults = gateways.get("default")
        if not defaults:
            return

        def default_ip(family):
            gw_info = defaults.get(family)
            if not gw_info:
                return
            addresses = netifaces.ifaddresses(gw_info[1]).get(family)
            if addresses:
                return addresses[0]["addr"]

        return default_ip(netifaces.AF_INET) or default_ip(netifaces.AF_INET6)
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def is_connected():
    if sys.platform == 'win32' or 'linux':
        try:
            socket.setdefaulttimeout(3)
            socket.socket(socket.AF_INET, socket.SOCK_STREAM).connect(("8.8.8.8", 53))
            return True
        except socket.error:
            return False
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_hostname():
    if sys.platform == 'win32' or 'linux':
        return socket.gethostname()
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_ping_time():
    if sys.platform == 'win32' or 'linux':
        st = speedtest.Speedtest()
        st.get_servers([])
        return str(st.results.ping) + 'ms'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_download_speed():
    if sys.platform == 'win32' or 'linux':
        wifi = speedtest.Speedtest()
        transfer = round(wifi.download())/1000000
        if transfer > 1000:
            return str(transfer/1000) + 'GBps'
        elif 1000 > transfer > 1:
            return str(transfer) + 'Mbps'
        elif transfer < 1:
            return str(transfer*1000) + 'Kbps'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()


def get_upload_speed():
    if sys.platform == 'win32' or 'linux':
        wifi = speedtest.Speedtest()
        transfer = round(wifi.upload()) / 1000000
        if transfer > 1000:
            return str(transfer / 1000) + 'GBps'
        elif 1000 > transfer > 1:
            return str(transfer) + 'Mbps'
        elif transfer < 1:
            return str(transfer * 1000) + 'Kbps'
    elif sys.platform == 'darwin':
        return unsupported_exception()
    else:
        return unsupported_exception()



import os
import sys
import sounddevice as sd
from core.exception import *


def get_audio_devices():
    if sys.platform == 'win32':
        return sd.query_devices()
    elif sys.platform == 'darwin':
        return
    elif sys.platform == 'linux':
        return os.popen('aplay -l').read()
    else:
        return unsupported_exception()



"""Unit tests for WMI modules

Some tests are optional, since they rely on remote machines and
usernames / passwords. To enable these, copy wmitest.master.ini
to wmitest.ini and set the parameters you have available.

The watcher tests spawn temporary processes and temporary
logical drives. These may get left behind.
"""

#
# TODO:
# - Test for negative timezone
# - Test for share name with embedded single quote
#

import os, sys
import datetime
try:
    import ConfigParser
except ImportError:
    import configparser as ConfigParser
import operator
try:
    import Queue
except ImportError:
    import queue as Queue
try:
    next
except NameError:
    def next(iterator): return iterator.next()
import subprocess
import tempfile
import threading
import time
import unittest
import warnings
try:
    import _winreg
except ImportError:
    import winreg as _winreg

import pythoncom
import win32api
import win32con
import win32file

import wmi

ini = ConfigParser.SafeConfigParser()
ini.read(["wmitest.master.ini", "wmitest.ini"])
settings = {}
if ini.has_section("settings"):
    settings.update(ini.items("settings"))
excludes = [i.strip() for i in settings.get("excludes", "").split(",")]

COMPUTERS = [None, "."]
if "machine" in settings:
    COMPUTERS.append(settings['machine'])
IMPERSONATION_LEVELS = [None, "identify", "impersonate", "delegate"]
AUTHENTICATION_LEVELS = [None, "default", "none", "connect", "call", "pkt", "pktintegrity", "pktprivacy"]
AUTHORITIES = [None]
if set(["domain", "machine"]) <= set(settings):
    #~ AUTHORITIES.append("kerberos:%s" % settings['domain'])
    AUTHORITIES.append("ntlmdomain:%s" % settings['domain'])
PRIVILEGES = [None, ['security', '!shutdown']]
NAMESPACES = [None, "root/cimv2", "default"]

class TestBasicConnections(unittest.TestCase):

    def test_basic_connection(self):
        "Check that a standard connection works"
        self.assert_(wmi.WMI())

    def test_remote_connection(self):
        "Check that a remote connection works, if specified"
        if "machine" in settings:
            self.assert_(wmi.WMI(settings['machine']))
        else:
            warnings.warn("Skipping test_remote_connection")

    def test_simple_moniker(self):
        "Check that a simple moniker works"
        self.assert_(wmi.WMI(moniker="winmgmts:"))

    def test_moniker_with_class(self):
        "Check that specifying a class in moniker works"
        c0 = wmi.WMI().Win32_ComputerSystem
        c1 = wmi.WMI(moniker="winmgmts:Win32_ComputerSystem")
        self.assert_(c0 == c1)

    def test_moniker_with_instance(self):
        "Check that specifying an instance in the moniker works"
        for c0 in wmi.WMI().Win32_ComputerSystem():
            break
        c1 = wmi.WMI(moniker='winmgmts:Win32_ComputerSystem.Name="%s"' % c0.Name)
        self.assert_(c0 == c1)

    def test_impersonation_levels(self):
        "Check that specifying an impersonation level works"
        for impersonation in IMPERSONATION_LEVELS:
            self.assert_(wmi.WMI(impersonation_level=impersonation))

    def test_authentication_levels(self):
        "Check that specifying an authentication level works"
        for authentication in AUTHENTICATION_LEVELS:
            try:
                c = wmi.WMI(authentication_level=authentication)
            except wmi.x_access_denied:
                warnings.warn("Access denied for authentication level %s" % authentication)
            else:
                self.assert_(c)

    def test_authority(self):
        "Check that specifying an authority works"
        for authority in AUTHORITIES:
            self.assert_(wmi.WMI(authority=authority))

    def test_privileges(self):
        "Check that specifying privileges works"
        for privileges in PRIVILEGES:
            self.assert_(wmi.WMI(privileges=privileges))

    def test_namespace(self):
        "Check that specifying a namespace works"
        for namespace in NAMESPACES:
            self.assert_(wmi.WMI(namespace=namespace))

    def test_suffix(self):
        "Check that a suffix returns the class of that name"
        self.assert_(wmi.WMI(namespace="DEFAULT", suffix="StdRegProv") == wmi.WMI(namespace="DEFAULT").StdRegProv)

    def test_user_password(self):
        "Check that username & password are passed through for a remote connection"
        if set(["machine", "user", "password"]) <= set(settings):
            self.assert_(wmi.WMI(computer=settings['machine'], user=settings['user'], password=settings['password']))
        else:
            warnings.warn("Skipping test_user_password because no machine, user or password")

    def test_too_much_authentication(self):
        "Check that user/password plus privs / suffix raises exception"
        self.assertRaises(wmi.x_wmi_authentication, wmi.WMI, computer='***', user="***", password="***", privileges=["***"])
        self.assertRaises(wmi.x_wmi_authentication, wmi.WMI, computer='***', user="***", password="***", suffix="***")

    def test_user_password_with_impersonation_level(self):
        "Check that an impersonation level works with a username / password"
        if not(set(["machine", "user", "password"]) <= set(settings)):
            warnings.warn("Skipping test_user_password_with_impersonation_level because no machine, user or password")
        else:
            self.assert_(
                wmi.WMI(
                    computer=settings['machine'],
                    user=settings['user'],
                    password=settings['password'],
                    impersonation_level="impersonate"
                )
            )

    def test_user_password_with_invalid_impersonation_level(self):
        "Check that an impersonation level works with a username / password"
        if not(set(["machine", "user", "password"]) <= set(settings)):
            warnings.warn("Skipping test_user_password_with_invalid_impersonation_level because no machine, user or password")
        else:
            self.assertRaises(
                wmi.x_wmi_authentication,
                wmi.WMI,
                computer=settings['machine'],
                user=settings['user'],
                password=settings['password'],
                impersonation_level="***"
            )

    def test_user_password_with_authentication_level(self):
        "Check that an invalid impersonation level raises x_wmi_authentication"
        if not(set(["machine", "user", "password"]) <= set(settings)):
            warnings.warn("Skipping test_user_password_with_authentication_level because no machine, user or password")
        else:
            self.assert_(
                wmi.WMI(
                    computer=settings['machine'],
                    user=settings['user'],
                    password=settings['password'],
                    authentication_level="pktIntegrity"
                )
            )

    def test_user_password_with_invalid_authentication_level(self):
        "Check that an invalid authentication level raises x_wmi_authentication"
        if not(set(["machine", "user", "password"]) <= set(settings)):
            warnings.warn("Skipping test_user_password_with_invalid_authentication_level because no machine, user or password")
        else:
            self.assertRaises(
                wmi.x_wmi_authentication,
                wmi.WMI,
                computer=settings['machine'],
                user=settings['user'],
                password=settings['password'],
                authentication_level="***"
            )

    def test_local_user_password(self):
        "Check that user/password for local connection raises exception"
        self.assertRaises(wmi.x_wmi_authentication, wmi.WMI, user="***", password="***")

    def test_find_classes(self):
        "Check ability to switch class scan on and off"
        self.assert_(wmi.WMI(find_classes=True)._classes)
        self.assertFalse(wmi.WMI(find_classes=False)._classes)

    def test_find_classes_false(self):
        "By default, don't scan for classes but load them on demand"
        self.assertFalse(wmi.WMI()._classes)
        self.assert_(wmi.WMI().classes)

    def test_classes_acts_as_list(self):
        self.assert_(wmi.WMI().classes.index)

    def test_classes_acts_as_dict(self):
        self.assert_(wmi.WMI().classes.keys)

class TestThreadedConnection(unittest.TestCase):


    def test_initialised_thread(self):
        """A WMI connection in a thread which has been initialised for COM
        should succeed.
        """
        def f(q):
            pythoncom.CoInitialize()
            try:
                try:
                    wmi.WMI()
                except:
                    q.put(False)
                else:
                    q.put(True)
            finally:
                pythoncom.CoUninitialize()

        q = Queue.Queue()
        threading.Thread(target=f, args=(q,)).start()
        self.assert_(q.get())

    def test_uninitialised_thread(self):
        """A WMI connection in a thread which has not been initialised
        for COM should fail with a wmi-specific exception.
        """
        def f(q):
            try:
                wmi.WMI()
            except wmi.x_wmi_uninitialised_thread:
                q.put(True)
            except:
                q.put(False)
            else:
                q.put(False)

        q = Queue.Queue()
        threading.Thread(target=f, args=(q,)).start()
        self.assert_(q.get())

class TestMoniker(unittest.TestCase):

    def test_moniker(self):
        """Look at all possible options for moniker construction and pass
        them through to a WMI connector
        """
        for computer in COMPUTERS:
            if computer in (None, "."):
                local_authorities = [None]
            else:
                local_authorities = AUTHORITIES
            for impersonation_level in IMPERSONATION_LEVELS:
                for authentication_level in AUTHENTICATION_LEVELS:
                    for authority in local_authorities:
                        for privileges in PRIVILEGES:
                            for namespace in NAMESPACES:
                                moniker = wmi.construct_moniker(
                                    computer=computer,
                                    impersonation_level=impersonation_level,
                                    authority=authority,
                                    privileges=privileges,
                                    namespace=namespace
                                )
                                self.assert_(wmi.WMI(moniker=moniker), "Moniker failed: %s" % moniker)

    def test_moniker_root_namespace(self):
        "Check that namespace is prefixed by root if needed"
        self.assertEquals(wmi.construct_moniker(namespace="default"), "winmgmts:root/default")
        self.assertEquals(wmi.construct_moniker(namespace="root/default"), "winmgmts:root/default")

class TestFunctions(unittest.TestCase):

    times = [
       ((2000, 1, 1), "20000101******.******+***"),
       ((2000, 1, 1, 10, 0, 0), "20000101100000.******+***"),
       ((2000, 1, 1, 10, 0, 0, 100), "20000101100000.000100+***"),
       ((2000, 1, 1, 10, 0, 0, 100, "GMT"), "20000101100000.000100+GMT")
    ]

    def test_signed_to_unsigned(self):
        tests = [
           (0, 0),
           (-1, 0xffffffff),
           (+1, 1),
           (0x7fffffff, 0x7fffffff),
           (-0x7fffffff, 0x80000001)
        ]
        for signed, unsigned in tests:
            self.assertEquals(wmi.signed_to_unsigned(signed), unsigned)

    def test_from_1601(self):
        "Check conversion from 100-ns intervals since 1601(!)"
        self.assertEquals(wmi.from_1601(0), datetime.datetime(1601, 1, 1))
        self.assertEquals(wmi.from_1601(24 * 60 * 60 * 10 * 1000 * 1000), datetime.datetime(1601, 1, 2))

    def test_from_time(self):
        "Check conversion from time-tuple to time-string"
        for t, s in self.times:
            self.assertEquals(wmi.from_time(*t), s)

    def test_to_time(self):
        "Check conversion from time-string to time-tuple"
        for t, s in self.times:
            t = tuple(list(t) +([None] * 8))[:8]
            self.assertEquals(wmi.to_time(s), t)

    def test_get_wmi_type(self):
        "Check that namespace, class & instance are identified correctly"
        self.assertEquals(wmi.get_wmi_type(wmi.WMI()), "namespace")
        self.assertEquals(wmi.get_wmi_type(wmi.WMI().Win32_ComputerSystem), "class")
        for i in wmi.WMI().Win32_ComputerSystem():
            self.assertEquals(wmi.get_wmi_type(i), "instance")

    def test_registry(self):
        """Convenience Registry function is identical to picking
        the StdRegProv class out of the DEFAULT namespace"""
        self.assertEquals(wmi.Registry(), wmi.WMI(namespace="DEFAULT").StdRegProv)

class TestWMI(unittest.TestCase):

    def setUp(self):
        self.connection = wmi.WMI(namespace="root/cimv2", find_classes=False)
        self.logical_disks = set(self.connection.Win32_LogicalDisk())

class TestNamespace(TestWMI):

    def test_subclasses_of_simple(self):
        self.assert_("Win32_ComputerSystem" in self.connection.subclasses_of())

    def test_subclasses_of_subtree(self):
        self.assert_("Win32_Desktop" in self.connection.subclasses_of("CIM_Setting"))

    def test_subclasses_of_pattern(self):
        self.assert_(set(["Win32_LogicalDisk", "Win32_MappedLogicalDisk"]) <= set(self.connection.subclasses_of("CIM_LogicalDevice", "Win32_.*Disk")))

    def test_instances(self):
        self.assertEquals(self.logical_disks, set(self.connection.instances("Win32_LogicalDisk")))

    def test_new(self):
        "Check this is an alias for the new method of the equivalent class"
        self.assertEquals(self.connection.new("Win32_Process")._instance_of, self.connection.Win32_Process)

    def test_query(self):
        self.assertEquals(self.logical_disks, set(self.connection.query("SELECT * FROM Win32_LogicalDisk")))

    def test_ipython_attributes_with_find_classes(self):
        connection = wmi.WMI(find_classes=True)
        self.assertEquals(sorted(connection._getAttributeNames()), sorted(i for i in connection.classes if not i.startswith("__")))

    def test_getattr(self):
        "Check that WMI classes are returned by attribute access on their namespace"
        connection = wmi.WMI(find_classes=True)
        for c in list(connection.classes)[:5]:
            wmi_class = getattr(connection, c)
            self.assert_(isinstance(wmi_class, wmi._wmi_class))
            self.assertEquals(wmi_class._class_name, c)

    def test_watch_for(self):
        """Check that the watch_for method returns a watcher. The watcher itself
        will be tested elsewhere.
        """
        watcher = self.connection.watch_for(
            wmi_class="Win32_Process"
        )
        self.assert_(isinstance(watcher, wmi._wmi_watcher))

class TestClass(TestWMI):

    def test_class_from_namespace(self):
        self.assert_(self.connection.Win32_ComputerSystem._namespace is self.connection)

    def test_class_without_namespace(self):
        wmi_class = wmi.GetObject("winmgmts:Win32_ComputerSystem")
        self.assert_(wmi._wmi_class(None, wmi_class)._namespace)

    def test_query(self):
        self.assertEquals(
            set(self.connection.Win32_ComputerSystem.query()),
            set(self.connection.query("SELECT * FROM Win32_ComputerSystem"))
        )

    def test_query_with_where(self):
        this_drive = os.getcwd()[:2]
        for drive in self.connection.Win32_LogicalDisk(Name=this_drive):
            self.assertEquals(drive.Name, this_drive)

    def test_query_with_fields(self):
        this_drive = os.getcwd()[:2]
        properties = set(["MediaType"])
        self.assert_("Name" not in properties)
        for drive in self.connection.Win32_LogicalDisk(properties, Name=this_drive):
            self.assertEquals(set(drive.properties), set(properties))
            self.assert_(drive.MediaType)
            self.assertRaises(AttributeError, getattr, drive, "Name")

    def test_watch_for(self):
        """Check that the watch_for method returns a watcher. The watcher itself
        will be tested elsewhere.
        """
        watcher = self.connection.Win32_Process.watch_for()
        self.assert_(isinstance(watcher, wmi._wmi_watcher))

    def test_instances(self):
        self.assertEquals(
            set(self.connection.Win32_LogicalDisk()),
            set(self.connection.Win32_LogicalDisk.instances())
        )

    def test_new(self):
        process = self.connection.Win32_Process.new()
        self.assertEquals(wmi.get_wmi_type(process), "instance")
        self.assertEquals(process._instance_of, self.connection.Win32_process)


class TestWatcher(TestWMI):

    def new_letter(self):
        return \
            set("%s:" % chr(i) for i in range(ord('A'), 1 + ord('Z'))).\
            difference(d.DeviceID for d in self.connection.Win32_LogicalDisk()).\
            pop()

    @staticmethod
    def create(new_letter):
        #~ print("about to create drive with letter", new_letter)
        here = os.path.dirname(os.path.abspath(__file__))
        win32file.DefineDosDevice(0, new_letter, here)
        try:
            #
            # This sleep is needed for the WMI pollster to react
            #
            time.sleep(2)
        finally:
            win32file.DefineDosDevice(2, new_letter, here)

    def test_creation(self):
        try:
            new_letter = self.new_letter()
        except KeyError:
            warnings.warn("Unable to find a spare drive letter to map.")
            return

        watcher = self.connection.Win32_LogicalDisk.watch_for(
            notification_type="Creation",
            DeviceID=new_letter
        )
        t = threading.Timer(2, self.create,(new_letter,))
        t.start()
        found_disk = watcher(timeout_ms=20000)
        self.assert_(isinstance(found_disk, wmi._wmi_object))
        self.assertEqual(found_disk.Caption, new_letter)
        t.join()

    def test_event_with_no_params(self):
        try:
            new_letter = self.new_letter()
        except KeyError:
            warnings.warn("Unable to find a spare drive letter to map.")
            return

        #
        # This watcher will return *any* logical disk with some activity. To
        # make sure there is at least some, we'll create a new logical disk
        # but there's no guarantee that this will be the one returned, as
        # activity on, eg, the C: drive will be enough to trigger the event.
        #
        watcher = self.connection.Win32_LogicalDisk.watch_for()
        t = threading.Timer(2, self.create,(new_letter,))
        t.start()
        found_disk = watcher(timeout_ms=20000)
        self.assert_(isinstance(found_disk, wmi._wmi_object))
        self.assertEqual(found_disk.path().Class, "Win32_LogicalDisk")
        t.join()

    def test_valid_notification_types(self):
        for notification_type in ['operation', 'modification', 'creation', 'deletion']:
            self.assert_(self.connection.Win32_LogicalDisk.watch_for(notification_type=notification_type))

    def test_invalid_notification_types(self):
        self.assertRaises(wmi.x_wmi, self.connection.Win32_LogicalDisk.watch_for, notification_type="***")


    def do_not_test_extrinsic_event(self):

        #
        # This doesn't seem implementable at the moment
        # as a test. I can't find a reproducible extrinsic
        # event except for Win32_DeviceChangeEvent and that
        # one would require someone to, eg, plug in / unplug
        # a USB stick.
        #
        # It looks as though Win32_ProcessStartTrace should work
        # and it does on my laptop; just not on my desktop.
        #

        def _create(queue):
            queue.put(subprocess.Popen([sys.executable, "-c", "import time; time.sleep(10)"]))

        watcher = self.connection.Win32_ProcessStartTrace.watch_for(
            fields=["*"]##,
            #~ ProcessName=os.path.basename(sys.executable)
        )
        q = Queue.Queue()
        t = threading.Timer(2, _create,(q,))
        try:
            t.start()
            found_process = watcher(timeout_ms=20000)
            spawned_process = q.get_nowait()
            self.assert_(isinstance(found_process, wmi._wmi_event))
            self.assertEqual(int(found_process.ProcessID), spawned_process.pid)
        finally:
            t.cancel()

class TestMethods(TestWMI):

    def test_exists(self):
        "Check that a well-known method is available by attribute"
        self.assert_(self.connection.Win32_Process.Create)

    def test_params(self):
        "Check that the names and arrayness of params are picked up when not arrays"
        self.assertEquals(
            [(n, False) for n in ["CommandLine", "CurrentDirectory", "ProcessStartupInformation"]],
            self.connection.Win32_Process.Create.in_parameter_names
        )
        self.assertEquals(
            [("ProcessId", False),("ReturnValue", False)],
            self.connection.Win32_Process.Create.out_parameter_names
        )

    def test_positional_params(self):
        dir = tempfile.mkdtemp()
        filename = "abc.txt"
        contents = str(datetime.datetime.now())
        handle, result = self.connection.Win32_Process.Create(
            "cmd /c echo %s > %s" %(contents, filename),
            dir,
            self.connection.Win32_ProcessStartup.new(ShowWindow=0)
        )
        time.sleep(0.5)
        with open(os.path.join(dir, filename)) as f:
            self.assertEqual(f.read(), contents + " \n")

    def test_named_params(self):
        dir = tempfile.mkdtemp()
        filename = "abc.txt"
        contents = str(datetime.datetime.now())
        handle, result = self.connection.Win32_Process.Create(
            ProcessStartupInformation=self.connection.Win32_ProcessStartup.new(ShowWindow=0),
            CurrentDirectory=dir,
            CommandLine="cmd /c echo %s > %s" %(contents, filename)
        )
        time.sleep(0.5)
        with open(os.path.join(dir, filename)) as f:
            self.assertEqual(f.read(), contents + " \n")

    def test_in_params_with_array(self):
        "Check that the names and arrayness of params are picked up when arrays"
        self.assertEquals(
            [("DNSServerSearchOrder", True)],
            self.connection.Win32_NetworkAdapterConfiguration.SetDNSServerSearchOrder.in_parameter_names
        )

    def test_instance_methods_are_distinct(self):
        """Check that the methods of difference instances of a class are distinct.
        This caused a problem when calling .Terminate on one process killed another.
        """
        methods = [d.Reset for d in self.logical_disks]
        for i in range(len(methods)-1):
            self.assertNotEqual(methods[i], methods[i+1])

    def test_call_from_class(self):
        "Check that a method can be called from a class"
        self.assert_(self.connection.Win32_Process.Create(
            CommandLine=sys.executable + " -c pass",
            ProcessStartupInformation=self.connection.Win32_ProcessStartup.new(ShowWindow=0)
        ))

    def test_call_from_instance(self):
        "Check that a method can be called from an instance"
        handle, _ = self.connection.Win32_Process.Create(
            CommandLine=sys.executable,
            ProcessStartupInformation=self.connection.Win32_ProcessStartup.new(ShowWindow=0)
        )
        result = 1
        for p in self.connection.Win32_Process(Handle=handle):
            result, = p.Terminate()
        self.assertEqual(result, 0)

class TestProperties(TestWMI):

    def test_access(self):
        "Check that all properties are available as attributes"
        for d in self.logical_disks:
            break
        for p in d.ole_object.Properties_:
            self.assertEqual(p.Value, getattr(d, p.Name))

    def test_attribute_passthrough(self):
        "Check that unknown attributes are passed through to the underlying object"
        for d in self.logical_disks:
            break
        #
        # Can't rely on the COM Objects testing identical or equal;
        # have to check their values and their emptiness.
        #
        self.assert_(d.Properties_)
        self.assert_(d.ole_object.Properties_)
        self.assertEqual(
            [p.Value for p in d.Properties_],
            [p.Value for p in d.ole_object.Properties_]
        )

    def test_settable(self):
        "Check that a writeable property can be written"
        name = str(time.time()).split(".")[0]
        old_value = "***"
        new_value = "!!!"
        username = win32api.GetUserNameEx(win32con.NameSamCompatible)
        self.assert_(not self.connection.Win32_Environment(Name=name, UserName=username))
        self.connection.Win32_Environment.new(Name=name, UserName=username, VariableValue=old_value).put()
        for envvar in self.connection.Win32_Environment(Name=name, UserName=username):
            self.assertEqual(envvar.VariableValue, old_value)
            envvar.VariableValue = new_value
        try:
            for envvar in self.connection.Win32_Environment(Name=name, UserName=username):
                self.assertEqual(envvar.VariableValue, new_value)
        finally:
            for envvar in self.connection.Win32_Environment(Name=name, UserName=username):
                envvar.VariableValue = None

class TestInstances(TestWMI):

    def test_hashable(self):
        "Ensure instances are hashable so can be used in a set/dict"
        self.assert_(dict.fromkeys(self.logical_disks))

    def test_equalable(self):
        "Ensure instances compare equal"
        self.assertEqual(self.logical_disks, self.logical_disks)

    def test_not_equal_to_anything_else(self):
        "Ensure WMI instances are not equal to non-WMI instances"
        for d in self.logical_disks:
            break
        self.assertNotEqual(d, d.Caption)

    def test_sortable(self):
        "Ensure instances sort by full path/key"
        self.assertEqual(
            sorted(self.logical_disks),
            sorted(self.logical_disks, key=operator.attrgetter("DeviceID"))
        )

    def test_references(self):
        "Ensure that associations are special-cased to return wrapped objects"
        for d in self.logical_disks:
            break
        for r in d.references("Win32_LogicalDiskRootDirectory"):
            self.assert_(r.is_association)
            self.assertEqual(r.GroupComponent, d)
            self.assert_(isinstance(r.GroupComponent, wmi._wmi_object))
            self.assert_(isinstance(r.PartComponent, wmi._wmi_object))

    def test_associators(self):
        "Ensure that associators are returned by association / result"
        for d in self.logical_disks:
            if d.DeviceID == os.path.abspath(__file__)[:2]:
                break
        else:
            raise RuntimeError("Unable to find the logical drive corresponding to this file")
        root_dir = d.associators(wmi_association_class="Win32_LogicalDiskRootDirectory")[0]
        self.assertEqual(root_dir.Name.lower(), d.Name.lower() + "\\".lower())
        root_dir = d.associators(wmi_result_class="Win32_Directory")[0]
        self.assertEqual(root_dir.Name.lower(), d.Name.lower() + "\\")

    def test_derivation(self):
        "Check that derivation mimics WMI-provided Derivation_ property"
        for d in self.logical_disks:
            break
        self.assertEqual(d.derivation(), d.ole_object.Derivation_)

    def test_keys(self):
        "Check that the readonly keys property returns the keys for an object"
        self.assertEqual(self.connection.Win32_LogicalDisk.keys, ['DeviceID'])
        self.assertEqual(next(iter(self.logical_disks)).keys, ['DeviceID'])

class TestInstanceCreation(TestWMI):

    def test_create_instance(self):
        self.assert_(isinstance(self.connection.Win32_ProcessStartup.new(ShowWindow=2), wmi._wmi_object))

class TestAssociations(TestWMI):

    def test_all_properties_available(self):
        #
        # An association can contain not only the associated
        # classes but also extra information as well. Ensure
        # that both types of data are correctly handled.
        #
        for q in self.connection.Win32_DiskQuota():
            for p in q.properties:
                try:
                    getattr(q, p)
                except wmi.x_wmi:
                    assert False, "Error getting %s from %s" % (p, q)
            else:
                assert True

if __name__ == '__main__':
    unittest.main()
